# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import six
import os.path

from ustorage.utils import get_random_string
from ustorage.utils import files


class Config(dict):
    '''
    Wrap the configuration for a single :class:`Storage`.
    Basically, it's an ObjectDict
    '''
    def __getattr__(self, name):
        if name in self:
            return self[name]
        else:
            raise AttributeError('Unknown attribute: ' + name)

    def __setattr__(self, name, value):
        self[name] = value

    def __delattr__(self, name):
        if name in self:
            del self[name]
        else:
            raise AttributeError('Unknown attribute: ' + name)


class BaseStorage(object):
    '''
    Abstract class to implement a storage backend.
    '''
    root = None
    DEFAULT_MIME = 'application/octet-stream'
    DEFAULT_CONFIG = {}

    def __init__(self, config):
        """
        Initialize a storage

        :param config: A dict-like configuration container.
        """
        self.config = Config(self.DEFAULT_CONFIG, **config)

    def exists(self, name):
        '''Test wether a file exists or not given its name in the storage'''
        raise NotImplementedError('Existance checking is not implemented')

    def open(self, name, *args, **kwargs):
        '''Open a file given its name relative to the storage root'''
        raise NotImplementedError('Open operation is not implemented')

    def read(self, name):
        '''Read a file content given its name in the storage'''
        raise NotImplementedError('Read operation is not implemented')

    def write(self, name, content):
        '''Write content into a file given its name in the storage'''
        raise NotImplementedError('Write operation is not implemented')

    def delete(self, name):
        '''Delete a file given its name in the storage'''
        raise NotImplementedError('Delete operation is not implemented')

    def copy(self, name, target):
        '''Copy a file given its name to another path in the storage'''
        raise NotImplementedError('Copy operation is not implemented')

    def move(self, name, target):
        '''
        Move a file given its name to another path in the storage

        Default implementation perform a copy then a delete.
        Backends should overwrite it if there is a better way.
        '''
        self.copy(name, target)
        self.delete(name)

    def save(self, file_or_wfs, name, overwrite=False):
        '''
        Save a file-like object or a `werkzeug.FileStorage` with the specified name.

        :param storage: The file or the storage to be saved.
        :param name: The destination in the storage.
        :param overwrite: if `False`, raise an exception if file exists in storage

        :raises FileExists: when file exists and overwrite is `False`
        '''
        self.write(name, file_or_wfs.read())
        return name

    def metadata(self, name):
        '''
        Fetch all available metadata for a given file
        '''
        meta = self.get_metadata(name)
        # Fix backend mime misdetection
        meta['mime'] = meta.get('mime') or files.mime(name, self.DEFAULT_MIME)
        return meta

    def get_metadata(self, name):
        '''
        Backend specific method to retrieve metadata for a given file
        '''
        raise NotImplementedError('Copy operation is not implemented')

    def as_binary(self, content, encoding='utf8'):
        '''Perform content encoding for binary write'''
        if hasattr(content, 'read'):
            return content.read()
        elif isinstance(content, six.text_type):
            return content.encode(encoding)
        else:
            return content

    def get_available_name(self, name):
        """
        Returns a name that's free on the target storage system, and
        available for new content to be written to.
        """
        dir_name, file_name = os.path.split(name)
        file_root, file_ext = os.path.splitext(file_name)
        # If the name already exists, add an underscore and a random 7
        # character alphanumeric string (before the file extension, if one
        # exists) to the name until the generated name doesn't exist.
        while self.exists(name):
            # file_ext includes the dot.
            name = os.path.join(
                dir_name,
                '{}_{}{}'.format(file_root, get_random_string(7), file_ext))
        return name



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import codecs
import tempfile
import io
import logging

from contextlib import contextmanager

import boto3

from botocore.exceptions import ClientError

from ustorage.bases import BaseStorage
from ustorage.utils import files
from ustorage.utils import drop_none_values
from ustorage.utils import CaseInsensitiveDict

log = logging.getLogger(__name__)


def _extract_metadata(obj):
    '''Extract metadata from s3 object'''
    checksum = 'md5:{0}'.format(obj.e_tag[1:-1])
    mime = obj.content_type.split(';', 1)[0] if obj.content_type else None
    meta = {
        'checksum': checksum,
        'size': obj.content_length,
        'mime': mime,
        'modified': obj.last_modified,
    }
    meta.update(obj.metadata)
    return meta


class S3Storage(BaseStorage):
    '''
    An Amazon S3 Backend (compatible with any S3-like API)

    Expect the following settings:

    - `endpoint`: The S3 API endpoint
    - `region`: The region to work on.
    - `access_key`: The AWS credential access key
    - `secret_key`: The AWS credential secret key
    '''
    DEFAULT_CONFIG = dict(
        endpoint=None,
        region=None,
        access_key=None,
        secret_key=None,
    )

    def __init__(self, config):
        super(S3Storage, self).__init__(config)
        config = self.config

        self._aws_session = aws_session = boto3.session.Session()
        self._s3_config = s3_config = boto3.session.Config(
            signature_version='s3v4'
        )

        self.s3 = aws_session.resource('s3',
                                       config=s3_config,
                                       endpoint_url=config.endpoint,
                                       region_name=config.region,
                                       aws_access_key_id=config.access_key,
                                       aws_secret_access_key=config.secret_key)
        self.bucket = self.s3.Bucket(config.bucket)
        self.backend_exceptions = self.s3.meta.client.exceptions

    def create_bucket(self):
        try:
            self.bucket.create()
        except self.s3.meta.client.exceptions.BucketAlreadyOwnedByYou:
            pass

    def exists(self, name):
        try:
            self.bucket.Object(name).load()
        except ClientError:
            return False
        return True

    def get(self, name):
        """Return a pair of body+metadata"""
        obj = self.bucket.Object(name)
        return (obj.get()['Body'].read(), _extract_metadata(obj))

    @contextmanager
    def open(self, name, mode='r', encoding='utf8'):
        """
        Returns a file-like object.

        Example:

            with storage.open('test.json', 'w') as f:
                json.dump({'a': 1}, f)

            with storage.open('test.json') as f:
                print(json.load(f))

        :param mode: gives "b" flag if of binary object,
                     gives "w" for writting.
        """
        obj = self.bucket.Object(name)
        if 'r' in mode:
            f = obj.get()['Body']
            yield f if 'b' in mode else codecs.getreader(encoding)(f)
        else:  # mode == 'w'
            f = tempfile.SpooledTemporaryFile()
            yield f
            f.seek(0)
            obj.put(Body=f.read())
            f.close()

    def read(self, name):
        obj = self.bucket.Object(name).get()
        return obj['Body'].read()

    def write(self, name, content, metadata=None):
        metadata = CaseInsensitiveDict(metadata)

        text_type = type('')
        for k, v in metadata.items():
            metadata[k] = text_type(v)

        if 'Content-Type' not in metadata:
            metadata.update({
                'Content-Type': files.mime(name)
            })

        return self.bucket.put_object(**drop_none_values(dict(
            Key=name,
            Body=self.as_binary(content),
            ContentType=metadata.pop('Content-Type', None),
            Metadata=dict(drop_none_values(metadata)),
        )))

    def delete(self, name):
        for obj in self.bucket.objects.filter(Prefix=name):
            obj.delete()

    def copy(self, name, target):
        src = {
            'Bucket': self.bucket.name,
            'Key': name,
        }
        self.bucket.copy(src, target)

    def list_files(self):
        for f in self.bucket.objects.all():
            yield f.key

    def get_metadata(self, name):
        '''Fetch all availabe metadata'''
        obj = self.bucket.Object(name)
        return _extract_metadata(obj)

    def get_url(self, name, parameters=None, expire=3600):
        params = parameters.copy() if parameters else {}
        params['Bucket'] = self.bucket.name
        params['Key'] = name
        return self.s3.meta.client.generate_presigned_url(
            'get_object', Params=params, ExpiresIn=expire
        )



import os
import re

from setuptools import setup, find_packages


def get_version():
    here = os.path.abspath(os.path.dirname(__file__))
    with open(os.path.join(here, 'ustorage', '__init__.py')) as f:
        return re.findall(r"^__version__ = '([^']+)'\r?$", f.read(), re.M)[0]


setup(
    name='ustorage',
    version=get_version(),
    description='Unified Storage Interface for Python.',
    license='Apache License 2.0',
    url='https://github.com/wonderbeyond/ustorage',
    author='wonderbeyond',
    author_email='wonderbeyond@gmail.com',
    packages=find_packages(exclude=['contrib', 'docs', 'tests*']),
    install_requires=[
        'six',
    ],
    extras_require={
        's3': 'boto3==1.*',
    },
    zip_safe=False,
)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import mimetypes
import os.path

__all__ = (
    'TEXT', 'DOCUMENTS', 'IMAGES', 'AUDIO', 'DATA', 'SCRIPTS', 'ARCHIVES', 'EXECUTABLES',
    'DEFAULTS', 'ALL', 'NONE', 'All', 'AllExcept', 'DisallowAll'
)

#: This just contains plain text files (.txt).
TEXT = ['txt']

#: This contains various office document formats
#: (.rtf, .odf, .ods, .gnumeric, .abw, .doc, .docx, .xls, and .xlsx).
#: Note that the macro-enabled versions of Microsoft Office 2007 files are not included.
DOCUMENTS = 'rtf odf ods gnumeric abw doc docx xls xlsx'.split()

#: This contains basic image types that are viewable from most browsers (.jpg,
#: .jpe, .jpeg, .png, .gif, .svg, and .bmp).
IMAGES = 'jpg jpe jpeg png gif svg bmp'.split()

#: This contains audio file types (.wav, .mp3, .aac, .ogg, .oga, and .flac).
AUDIO = 'wav mp3 aac ogg oga flac'.split()

#: This is for structured data files (.csv, .ini, .json, .plist, .xml, .yaml, and .yml).
DATA = 'csv ini json plist xml yaml yml'.split()

#: This contains various types of scripts (.js, .php, .pl, .py .rb, and .sh).
#: If your Web server has PHP installed and set to auto-run,
#: you might want to add ``php`` to the DENY setting.
SCRIPTS = 'js php pl py rb sh bat'.split()

#: This contains archive and compression formats (.gz, .bz2, .zip, .tar, .tgz, .txz, and .7z).
ARCHIVES = 'gz bz2 zip tar tgz txz 7z'.split()

#: This contains shared libraries and executable files (.so, .exe and .dll).
#: Most of the time, you will not want to allow this - it's better suited for use with `AllExcept`.
EXECUTABLES = 'so exe dll'.split()

#: The default allowed extensions - `TEXT`, `DOCUMENTS`, `DATA`, and `IMAGES`.
DEFAULTS = TEXT + DOCUMENTS + IMAGES + DATA


def extension(filename):
    ext = os.path.splitext(filename)[1]
    if ext.startswith('.'):
        # os.path.splitext retains . separator
        ext = ext[1:]
    return ext.lower()


def lower_extension(filename):
    '''
    This is a helper used by :meth:`Storage.save` to provide lowercase extensions for
    all processed files, to compare with configured extensions in the same
    case.

    :param str filename: The filename to ensure has a lowercase extension.
    '''
    if '.' in filename:
        main, ext = os.path.splitext(filename)
        return main + ext.lower()
    # For consistency with os.path.splitext,
    # do not treat a filename without an extension as an extension.
    # That is, do not return filename.lower().
    return filename


def mime(filename, default=None):
    '''
    A basic helper to guess mime type from a filename or url
    '''
    return mimetypes.guess_type(filename)[0] or default


class All(object):
    '''
    This type can be used to allow all extensions.
    There is a predefined instance named `ALL`.
    '''
    def __contains__(self, item):
        return True


class DisallowAll(object):
    '''
    This type can be used to disallow all extensions.
    There is a predefined instance named `NONE`.
    '''
    def __contains__(self, item):
        return False


#: This "contains" all items. You can use it to allow all extensions to be uploaded.
ALL = All()

#: This "contains" no items. You can use it to force a whitelist by configuration.
NONE = DisallowAll()


class AllExcept(object):
    '''
    This can be used to allow all file types except certain ones.

    For example, to exclude .exe and .iso files, pass::

        AllExcept(('exe', 'iso'))

    to the :class:`~flask_fs.Storage` constructor as `extensions` parameter.

    You can use any container, for example::

        AllExcept(SCRIPTS + EXECUTABLES)
    '''
    def __init__(self, items):
        self.items = items

    def __contains__(self, item):
        return item not in self.items
